import logging
import time
from typing import Any, Dict, List, Optional

from fog_x.database import DatabaseManager
from fog_x.feature import FeatureType

logger = logging.getLogger(__name__)


class Episode:
    def __init__(
        self,
        db_manager: DatabaseManager,
        metadata: Optional[Dict[str, Any]] = {},
        features: Dict[str, FeatureType] = {},
        enable_feature_inference=True,  # whether additional FeatureTypes can be inferred
    ):
        self.features = features
        self.enable_feature_inference = enable_feature_inference
        self.db_manager = db_manager
        self.db_manager.initialize_episode(additional_metadata=metadata)

    def add(
        self,
        feature: str,
        value: Any,
        timestamp: Optional[int] = None,
        feature_type: Optional[FeatureType] = None,
        metadata_only = False,
    ) -> None:
        """
        Add one feature step data.
        To add multiple features at each step, call this multiple times or use
        `add_by_dict` to ensure the same timestamp is used for each feature.

        Args:
            feature (str): name of the feature
            value (Any): value associated with the feature
            timestamp (optional int): nanoseconds since the Epoch.
                If not provided, the current time is used.

        Examples:
            >>> episode.add('feature1', 'image1.jpg')
            >>> episode.add('feature2', 100)
        """

        if timestamp is None:
            timestamp = time.time_ns()

        logger.debug(
            f"Adding {feature} with value {value} at timestamp {timestamp}"
        )
        if self.db_manager:
            self.db_manager.add(feature, value, timestamp, feature_type, metadata_only)
        else:
            logger.warning(
                "No database manager provided, data will not be saved"
            )

    def add_by_dict(
        self, data: Dict[str, Any], timestamp: Optional[int] = None
    ) -> None:
        """
        Add multiple features as step data.
        Ensures that the same timestamp is used for each feature.

        Args:
            data (Dict[str, Any]): feature names and their values
            timestamp (optional int): nanoseconds since the Epoch.
                If not provided, the current time is used.

        Examples:
            >>> episode.add_by_dict({'feature1': 'image1.jpg', 'feature2': 100})
        """
        if timestamp is None:
            timestamp = time.time_ns()
        for feature, value in data.items():
            self.add(feature, value, timestamp)

    def compact(self) -> None:
        """
        Creates a table for the compacted data.

        TODO:
            * compact should not be run more than once?
            * expand docstring description
        """
        self.db_manager.compact()

    def get_steps(self) -> List[Dict[str, Any]]:
        """
        Retrieves the episode's step data.

        Returns:
            the step data

        TODO:
            * get_steps not in db_manager; db_manager.get_step_table_all returns a `LazyFrame`, not `List[Dict[str, Any]]`
        """
        return self.db_manager.get_steps()

    def close(self, save_data = True, save_metadata = True, additional_metadata = {}) -> None:
        """
        Saves the episode object.
        """
        self.db_manager.close(save_data=save_data, save_metadata=save_metadata, additional_metadata=additional_metadata)



import logging
from typing import Any, List

import pandas as pd  # type: ignore
import sqlalchemy  # type: ignore
from sqlalchemy import MetaData  # type: ignore
from sqlalchemy import Table  # type: ignore
from sqlalchemy import text  # type: ignore
from sqlalchemy import Column, Integer, create_engine, inspect
from sqlalchemy.orm import declarative_base, sessionmaker  # type: ignore

from fog_x.database.utils import type_py2sql  # type: ignore

Base = declarative_base()
logger = logging.getLogger(__name__)


class DatabaseConnector:
    def __init__(self, path: str):
        self.engine = create_engine(f"sqlite:///{path}")
        Base.metadata.bind = self.engine
        DBSession = sessionmaker(bind=self.engine)
        self.session = DBSession()

    def add(self, key, value):
        # This method should be updated based on the specific use case.
        pass

    def query(self, query):
        # This should be updated to use SQLAlchemy's query capabilities.
        pass

    def close(self):
        self.session.close()

    def list_tables(self):
        inspector = inspect(self.engine)
        return inspector.get_table_names()

    def create_table(self, table_name: str, columns):
        metadata = MetaData()
        columns = [Column("id", Integer, primary_key=True)] + [
            Column(column_name, column_type)
            for column_name, column_type in columns.items()
        ]
        table = Table(table_name, metadata, *columns)
        metadata.create_all(self.engine)
        logger.info(f"Table {table_name} created.")

    def add_column(self, table_name, column):
        column_name = column.compile(dialect=self.engine.dialect)
        column_type = column.type.compile(self.engine.dialect)
        sql = text(
            f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}"
        )
        self.engine.execute(sql)
        # self.engine.execute('ALTER TABLE %s ADD COLUMN %s %s' % (table_name, column_name, column_type))

    def insert_data(
        self,
        table_name: str,
        data: dict,
        create_new_column_if_not_exist: bool = False,
    ) -> int:
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=self.engine)

        logger.info(f"Inserting data into {table_name} with data {data}")
        # Check for each key in data if it exists as a column
        if create_new_column_if_not_exist:
            for key, value in data.items():
                if table.c.get(key) is None:
                    logger.warn(
                        f"Creating new column {key} in table {table_name} with type {type(value)}"
                    )
                    self.add_column(
                        table_name,
                        Column(key, type_py2sql(type(value)), nullable=True),
                    )

        insert_result = self.engine.execute(table.insert(), data)
        logger.warn(
            f"Data inserted into {table_name} with index {insert_result.inserted_primary_key[0]}"
        )
        return insert_result.inserted_primary_key[0]

    def update_data(
        self,
        table_name: str,
        index: int,
        data: dict,
        create_new_column_if_not_exist: bool = False,
        is_partial_data: bool = False,
    ):
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=self.engine)

        logger.info(f"Inserting data into {table_name} with data {data}")
        # Check for each key in data if it exists as a column
        if create_new_column_if_not_exist:
            for key, value in data.items():
                if table.c.get(key) is None:
                    logger.warn(
                        f"Creating new column {key} in table {table_name} with type {type(value)}"
                    )
                    # self.add_column(
                    #     table_name,
                    #     Column(key, type_py2sql(type(value)), nullable=True),
                    # )
                    column_type = type_py2sql(type(value))
                    self.add_column(
                        table_name, Column(key, column_type, nullable=True)
                    )
                    metadata.clear()
                    table = Table(
                        table_name, metadata, autoload_with=self.engine
                    )
                    logger.info(
                        f"Successfully added column {key} to {table_name}"
                    )

        # if is_partial_data:
        #     self.engine.execute(
        #         table.update().where(table.c.id == index).values(**data)
        #     )
        #     logger.info(data)

        self.engine.execute(
            table.update().where(table.c.id == index).values(data)
        )
        logger.debug(f"Data updated in {table_name} at index {index}")

    def merge_tables_with_timestamp(
        self, tables: List[str], output_table: str
    ):
        merged_df = pd.DataFrame()
        # This method should be updated based on the specific use case.
        for table in tables:
            if merged_df.empty:
                merged_df = pd.read_sql(
                    f"SELECT * FROM {table}", self.engine
                ).drop("id", axis=1)
            else:
                cur_df = pd.read_sql(
                    f"SELECT * FROM {table}", self.engine
                ).drop("id", axis=1)
                merged_df = pd.merge_asof(
                    cur_df, merged_df, on="Timestamp", direction="nearest"
                )
                logger.info(f"merged: {merged_df}")
        logger.warn(
            """currently using merge asof nearest policy on the timstamp, 
            better policies should be expected"""
        )
        # policy can be X fps timestamp, and use backward merge_asof
        merged_df.to_sql(output_table, self.engine, if_exists="replace")

    def select_table(self, table_name: str, format: str = "sql") -> Any:
        if format == "sql":
            table = Table(table_name, MetaData(), autoload_with=self.engine)
            return self.engine.execute(
                sqlalchemy.select([table])
            )  # TODO: this may not work with the connection
        elif format == "pandas":
            return pd.read_sql(f"SELECT * FROM {table_name}", self.engine)
        else:
            raise ValueError(f"Unsupported table select format {format}")



import datetime
import decimal

import numpy as np
import pyarrow as pa
import sqlalchemy  # type: ignore
from polars import datatypes as pld

_type_py2sql_dict = {
    int: sqlalchemy.sql.sqltypes.BigInteger,
    str: sqlalchemy.sql.sqltypes.Unicode,
    float: sqlalchemy.sql.sqltypes.Float,
    decimal.Decimal: sqlalchemy.sql.sqltypes.Numeric,
    datetime.datetime: sqlalchemy.sql.sqltypes.DateTime,
    bytes: sqlalchemy.sql.sqltypes.LargeBinary,
    bool: sqlalchemy.sql.sqltypes.Boolean,
    datetime.date: sqlalchemy.sql.sqltypes.Date,
    datetime.time: sqlalchemy.sql.sqltypes.Time,
    datetime.timedelta: sqlalchemy.sql.sqltypes.Interval,
    list: sqlalchemy.sql.sqltypes.ARRAY,
    dict: sqlalchemy.sql.sqltypes.JSON,
}


def type_py2sql(pytype):
    """Return the closest sql type for a given python type"""
    if pytype in _type_py2sql_dict:
        return _type_py2sql_dict[pytype]
    else:
        raise NotImplementedError(
            "You may add custom `sqltype` to `"
            + str(pytype)
            + "` assignment in `_type_py2sql_dict`."
        )


def type_np2py(dtype=None, arr=None):
    """Return the closest python type for a given numpy dtype"""

    if (dtype is None and arr is None) or (
        dtype is not None and arr is not None
    ):
        raise ValueError(
            "Provide either keyword argument `dtype` or `arr`: a numpy dtype or a numpy array."
        )

    if dtype is None:
        dtype = arr.dtype

    # 1) Make a single-entry numpy array of the same dtype
    # 2) force the array into a python 'object' dtype
    # 3) the array entry should now be the closest python type
    single_entry = np.empty([1], dtype=dtype).astype(object)

    return type(single_entry[0])


def type_np2sql(dtype=None, arr=None):
    """Return the closest sql type for a given numpy dtype"""
    return type_py2sql(type_np2py(dtype=dtype, arr=arr))


def _datasets_dtype_to_arrow(datasets_dtype: str) -> pa.DataType:
    """
    _datasets_dtype_to_arrow takes a datasets string dtype and converts it to a pyarrow.DataType.
    In effect, `dt == datasets_dtype_to_arrow(_datasets_dtype_to_arrow(dt))`
    """

    if datasets_dtype == "null":
        return pa.null()
    elif datasets_dtype == "bool":
        return pa.bool_()
    elif datasets_dtype == "int8":
        return pa.int8()
    elif datasets_dtype == "int16":
        return pa.int16()
    elif datasets_dtype == "int32":
        return pa.int32()
    elif datasets_dtype == "int64":
        return pa.int64()
    elif datasets_dtype == "uint8":
        return pa.uint8()
    elif datasets_dtype == "uint16":
        return pa.uint16()
    elif datasets_dtype == "uint32":
        return pa.uint32()
    elif datasets_dtype == "uint64":
        return pa.uint64()
    elif datasets_dtype == "float16":
        return pa.float16()
    elif datasets_dtype == "float32":
        return pa.float32()
    elif datasets_dtype == "float64":
        return pa.float64()
    elif datasets_dtype == "binary":
        return pa.binary()
    elif datasets_dtype == "large_binary":
        return pa.large_binary()
    elif datasets_dtype == "string":
        return pa.string()
    elif datasets_dtype == "str":
        return pa.string()
    elif datasets_dtype == "large_string":
        return pa.large_string()
    elif datasets_dtype.startswith("timestamp"):
        # Extracting unit and timezone from the string
        unit, tz = None, None
        if "[" in datasets_dtype and "]" in datasets_dtype:
            unit = datasets_dtype[
                datasets_dtype.find("[") + 1 : datasets_dtype.find("]")
            ]
            tz_start = datasets_dtype.find("tz=")
            if tz_start != -1:
                tz = datasets_dtype[tz_start + 3 :]
        return pa.timestamp(unit, tz)
    else:
        raise ValueError(
            f"Datasets dtype {datasets_dtype} does not have a PyArrow dtype equivalent."
        )


def _datasets_dtype_to_pld(datasets_dtype: str) -> pld.DataType:
    """
    _datasets_dtype_to_pld takes a datasets string dtype and converts it to a polars.DataType.
    In effect, `dt == _datasets_dtype_to_pld(_pld_to_datasets_dtype(dt))` (assuming _pld_to_datasets_dtype is the reverse mapping).
    """

    if datasets_dtype == "null":
        return pld.Null
    elif datasets_dtype == "bool":
        return pld.Boolean
    elif datasets_dtype == "int8":
        return pld.Int8
    elif datasets_dtype == "int16":
        return pld.Int16
    elif datasets_dtype == "int32":
        return pld.Int32
    elif datasets_dtype == "int64":
        return pld.Int64
    elif datasets_dtype == "uint8":
        return pld.UInt8
    elif datasets_dtype == "uint16":
        return pld.UInt16
    elif datasets_dtype == "uint32":
        return pld.UInt32
    elif datasets_dtype == "uint64":
        return pld.UInt64
    elif datasets_dtype == "float16":
        # Note: Polars does not have a native float16 type. This will be mapped to float32.
        return pld.Float32
    elif datasets_dtype == "float32":
        return pld.Float32
    elif datasets_dtype == "float64":
        return pld.Float64
    elif datasets_dtype == "binary":
        return pld.Binary
    elif datasets_dtype == "large_binary":
        # Polars does not differentiate between binary and large_binary
        return pld.Binary
    elif (
        datasets_dtype == "string"
        or datasets_dtype == "str"
        or datasets_dtype == "large_string"
    ):
        # Polars treats all strings as Utf8
        return pld.Utf8
    elif datasets_dtype == "object":
        return pld.Object
    elif datasets_dtype.startswith("timestamp"):
        # Handling timestamps in Polars is a bit different. You might need additional handling based on your use case.
        return pld.Datetime
    else:
        raise ValueError(
            f"Datasets dtype {datasets_dtype} does not have a Polars dtype equivalent."
        )



import logging
import os
from typing import List

import polars as pl
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq

from fog_x.database.utils import _datasets_dtype_to_pld
import smart_open
logger = logging.getLogger(__name__)


class PolarsConnector:
    def __init__(self, path: str):
        # In Polars, data is directly read into DataFrames from files, not through a database engine
        self.path = path
        self.tables = (
            {}
        )  # This will store table names as keys and DataFrames as values
        self.table_len = {}

    def list_tables(self):
        # Listing available DataFrame tables
        return list(self.tables.keys())

    def add_column(self, table_name: str, column_name: str, column_type):
        if column_name in self.tables[table_name].columns:
            logger.debug(
                f"Column {column_name} already exists in table {table_name}."
            )
            return
        # Add a new column to an existing DataFrame
        if table_name in self.tables:
            arrow_type = _datasets_dtype_to_pld(column_type)
            # self.tables[table_name] = self.tables[table_name].with_column(pl.lit(None).alias(column_name).cast(column_type))
            # self.tables[table_name] = self.tables[table_name].with_columns(pl.lit(None).alias(column_name).cast(arrow_type))
            self.tables[table_name] = self.tables[table_name].with_columns(
                pl.Series(
                    column_name, [None] * self.table_len[table_name]
                ).cast(arrow_type)
            )
            logger.debug(f"Column {column_name} added to table {table_name}.")
        else:
            logger.error(f"Table {table_name} does not exist.")

    def insert_data(self, table_name: str, data: dict):
        # Inserting a new row into the DataFrame and return the index of the new row
        if table_name in self.tables:
            # use the schema of the original table
            new_row = pl.DataFrame(
                [data], schema=self.tables[table_name].schema
            )

            index_of_new_row = self.table_len[table_name]
            logger.debug(
                f"Inserting data into {table_name}: {data} with {new_row} to table {self.tables[table_name]}"
            )
            self.tables[table_name] = pl.concat(
                [self.tables[table_name], new_row], how="align"
            )
            # logger.debug(f"table is now {self.tables[table_name]}")

            self.table_len[table_name] += 1

            return index_of_new_row  # Return the index of the inserted row
        else:
            logger.error(f"Table {table_name} does not exist.")
            return None  # Or raise an exception depending on your error handling strategy

    def update_data(self, table_name: str, index: int, data: dict):
        # update data
        for column_name, value in data.items():
            logger.debug(
                f"updating {column_name} with {value} at index {index}"
            )
            self.tables[table_name][index, column_name] = value

    def merge_tables_with_timestamp(
        self, tables: List[str], output_table: str, 
        clear_feature_tables: bool = False
    ):
        for table_name in self.tables.keys():
            if table_name not in tables:
                continue
            self.tables[table_name] = self.tables[table_name].set_sorted(
                "Timestamp"
            )

        # Merging tables using timestamps
        if len(tables) > 1:
            merged_df = self.tables[tables[0]].join_asof(
                self.tables[tables[1]].drop("episode_id"), on="Timestamp", strategy="nearest"
            )
            for table_name in tables[2:]:
                merged_df = merged_df.join_asof(
                    self.tables[table_name].drop("episode_id"),
                    on="Timestamp",
                    strategy="nearest",
                )
            logger.debug("Tables merged on Timestamp.")
        else:
            merged_df = self.tables[tables[0]]
            logger.debug("Only one table to merge.")
        self.tables[output_table] = merged_df

    def remove_tables(self, tables: List[str]):
        for table_name in tables:
            self.tables.pop(table_name)
            logger.debug(f"Table {table_name} removed.")

    def select_table(self, table_name: str):
        # Return the DataFrame
        if table_name in self.tables:
            return self.tables[table_name]
        else:
            logger.error(
                f"Table {table_name} does not exist, available tables are {self.tables.keys()}."
            )
            return None


class DataFrameConnector(PolarsConnector):
    def __init__(self, path: str):
        super().__init__(path)
        self.tables = {}
        self.table_len = {}

    def create_table(self, table_name: str):
        # Create a new DataFrame with specified columns
        # schema = {column_name: _datasets_dtype_to_arrow(column_type) for column_name, column_type in columns.items()}
        self.tables[table_name] = pl.DataFrame()
        logger.debug(f"Table {table_name} created with Polars.")
        self.table_len[table_name] = 0  # no entries yet


    def load_tables(self, table_names: List[str]):
        
        # load tables from the path
        for table_name in table_names:
            path = self.path.removesuffix("/")
            path = f"{path}/{table_name}.parquet"
            logger.info(f"Prepare to load table {table_name} loaded from {path}.")

            # use smart_open to handle different file systems
            path = os.path.expanduser(path)
            try:
                with smart_open.open(path):
                    self.tables[table_name] = pl.from_arrow(pq.read_table(path))
                    self.table_len[table_name] = len(self.tables[table_name])
                    logger.info(f"Table {table_name} loaded from {path}.")
            except Exception as e:
                logger.warn(f"Failed to load table {table_name} from {path}.")
                continue 



    def save_table(self, table_name: str):
        pq.write_table(self.tables[table_name].to_arrow(), f"{self.path}/{table_name}.parquet")


class LazyFrameConnector(PolarsConnector):
    def __init__(self, path: str):
        super().__init__(path)
        self.tables = {}
        self.table_len = {}
        self.dataset = pl.scan_pyarrow_dataset(
            ds.dataset(self.path, format="parquet")
        )

    def load_tables(self, episode_ids: List[str], table_names: List[str]):
        # rescan the dataset
        self.dataset = pl.scan_pyarrow_dataset(
            ds.dataset(self.path, format="parquet")
        )

        for table_name, episode_id in zip(table_names, episode_ids):
            self.tables[table_name] = self.dataset.filter(
                pl.col("episode_id") == episode_id
            )
            # logger.debug(f"Tables loaded from {self.tables}")

    def create_table(self, table_name: str):
        # Create a new DataFrame with specified columns
        # schema = {column_name: _datasets_dtype_to_arrow(column_type) for column_name, column_type in columns.items()}
        self.tables[table_name] = pl.DataFrame()
        logger.debug(f"Table {table_name} created with Polars.")
        self.table_len[table_name] = 0  # no entries yet

    def save_table(self, table_name: str):
        # for table_name in tables:
        #     table = self.tables[table_name].to_arrow()
        #     pq.write_to_dataset(table, root_path=self.path)
        dataset = self.tables[table_name].to_arrow()
        # dataset = [self.dataset.to_table(), ]
        basename_template = f"{table_name}-{{i}}.parquet"
        ds.write_dataset(
            dataset,
            base_dir=self.path,
            basename_template=basename_template,
            format="parquet",
            existing_data_behavior="overwrite_or_ignore",
        )

    def get_dataset_table(self, reload:bool = False):
        if reload:
            self.dataset = pl.scan_pyarrow_dataset(
                ds.dataset(self.path, format="parquet")
            )
        return self.dataset


import os

__root_dir__ = os.path.dirname(os.path.abspath(__file__))


from fog_x import dataset, episode, feature
from fog_x.dataset import Dataset

all = ["dataset", "feature", "episode", "Dataset"]

import logging

_FORMAT = "%(levelname).1s %(asctime)s %(filename)s:%(lineno)d] %(message)s"
logging.basicConfig(format=_FORMAT)
logging.root.setLevel(logging.INFO)



import numpy as np
import tensorflow_datasets as tfds  # type: ignore
from IPython import display
from PIL import Image

DATASETS = [
    "fractal20220817_data",
    "kuka",
    "bridge",
    "taco_play",
    "jaco_play",
    "berkeley_cable_routing",
    "roboturk",
    "nyu_door_opening_surprising_effectiveness",
    "viola",
    "berkeley_autolab_ur5",
    "toto",
    "language_table",
    "columbia_cairlab_pusht_real",
    "stanford_kuka_multimodal_dataset_converted_externally_to_rlds",
    "nyu_rot_dataset_converted_externally_to_rlds",
    "stanford_hydra_dataset_converted_externally_to_rlds",
    "austin_buds_dataset_converted_externally_to_rlds",
    "nyu_franka_play_dataset_converted_externally_to_rlds",
    "maniskill_dataset_converted_externally_to_rlds",
    "cmu_franka_exploration_dataset_converted_externally_to_rlds",
    "ucsd_kitchen_dataset_converted_externally_to_rlds",
    "ucsd_pick_and_place_dataset_converted_externally_to_rlds",
    "austin_sailor_dataset_converted_externally_to_rlds",
    "austin_sirius_dataset_converted_externally_to_rlds",
    "bc_z",
    "usc_cloth_sim_converted_externally_to_rlds",
    "utokyo_pr2_opening_fridge_converted_externally_to_rlds",
    "utokyo_pr2_tabletop_manipulation_converted_externally_to_rlds",
    "utokyo_saytap_converted_externally_to_rlds",
    "utokyo_xarm_pick_and_place_converted_externally_to_rlds",
    "utokyo_xarm_bimanual_converted_externally_to_rlds",
    "robo_net",
    "berkeley_mvp_converted_externally_to_rlds",
    "berkeley_rpt_converted_externally_to_rlds",
    "kaist_nonprehensile_converted_externally_to_rlds",
    "stanford_mask_vit_converted_externally_to_rlds",
    "tokyo_u_lsmo_converted_externally_to_rlds",
    "dlr_sara_pour_converted_externally_to_rlds",
    "dlr_sara_grid_clamp_converted_externally_to_rlds",
    "dlr_edan_shared_control_converted_externally_to_rlds",
    "asu_table_top_converted_externally_to_rlds",
    "stanford_robocook_converted_externally_to_rlds",
    "eth_agent_affordances",
    "imperialcollege_sawyer_wrist_cam",
    "iamlab_cmu_pickup_insert_converted_externally_to_rlds",
    "uiuc_d3field",
    "utaustin_mutex",
    "berkeley_fanuc_manipulation",
    "cmu_play_fusion",
    "cmu_stretch",
    "berkeley_gnm_recon",
    "berkeley_gnm_cory_hall",
    "berkeley_gnm_sac_son",
]


def dataset2path(dataset_name):
    if dataset_name == "robo_net":
        version = "1.0.0"
    elif dataset_name == "language_table":
        version = "0.0.1"
    else:
        version = "0.1.0"
    return f"gs://gresearch/robotics/{dataset_name}/{version}"


def as_gif(images, path="temp.gif"):
    # Render the images as the gif:
    images[0].save(
        path, save_all=True, append_images=images[1:], duration=1000, loop=0
    )
    gif_bytes = open(path, "rb").read()
    return gif_bytes


def get_dataset_info(datasets):
    """
    Get information about the datasets.

    Args:
      datasets (list): List of dataset names.

    Returns:
      list: List of tuples containing dataset name and dataset information.
    """
    ret = []
    for name in datasets:
        uri = dataset2path(name)
        b = tfds.builder_from_directory(builder_dir=uri)
        split = list(b.info.splits.keys())[0]
        b.as_dataset(split=split)
        ret.append((name, b.info))
    return ret



from fog_x.database.db_connector import DatabaseConnector
from fog_x.database.db_manager import DatabaseManager
from fog_x.database.polars_connector import (
    DataFrameConnector,
    LazyFrameConnector,
    PolarsConnector,
)

# from fog_x.db.postgres import Postgres

__all__ = [
    "DatabaseConnector",
    "DatabaseManager",
    "PolarsConnector",
    "LazyFrameConnector",
]



import sys
import pytest

# import fog_x


def test_import():
    # each test runs on cwd to its temp dir
    import fog_x


def test_dataset_create():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )


def test_episode_create():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )
    trajectory = dataset.new_episode()
    trajectory.add(feature="hello", value=1.0)
    trajectory.add(feature="world", value=2.0)
    trajectory.close()


def test_dataset_read():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )
    for episode in dataset.read_by(
        pandas_metadata=dataset.get_metadata_as_pandas_df()
    ):
        print(episode)


def test_dataset_export():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )
    dataset.export(
        "/tmp/test_fog_x_export",
        format="rtx",
        obs_keys=["hello"],
        act_keys=["world"],
    )


def test_rtx_example_load():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )

    dataset.load_rtx_episodes(
        name="berkeley_autolab_ur5",
        split="train[:1]",
    )

    dataset.export("/tmp/rtx_export", format="rtx")


def test_rtx_example_merge():
    import fog_x

    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )
    dataset.load_rtx_episodes(
        name="berkeley_autolab_ur5",
        split="train[:2]",
        additional_metadata={
            "collector": "User 1",
            "custom_tag": "Partition_1",
        },
    )

    dataset.load_rtx_episodes(
        name="berkeley_autolab_ur5",
        split="train[3:5]",
        additional_metadata={
            "collector": "User 2",
            "custom_tag": "Partition_2",
        },
    )


def test_rtx_example_query():
    dataset = fog_x.Dataset(
        name="test_fog_x",
        path="/tmp/test_fog_x",
    )
    metadata = dataset.get_metadata_as_pandas_df()
    print(metadata)
    metadata = metadata.filter(metadata["custom_tag"] == "Partition_1")
    episodes = dataset.read_by(metadata)
    for episode in episodes:
        print(episode)



import io
import logging
import os
from typing import Any, Dict, List, Optional, Tuple
import subprocess
import numpy as np
import polars
import pandas

from fog_x.database import (
    DatabaseConnector,
    DatabaseManager,
    DataFrameConnector,
    LazyFrameConnector,
    PolarsConnector,
)
from fog_x.episode import Episode
from fog_x.feature import FeatureType

logger = logging.getLogger(__name__)



def convert_to_h264(input_file, output_file):
    
    # FFmpeg command to convert video to H.264
    command = [
        'ffmpeg',
        '-i', input_file,    # Input file
        '-loglevel', 'error', # Suppress the logs
        '-vcodec', 'h264', # Specify the codec
        output_file          # Output file
    ]
    subprocess.run(command)

def create_cloud_bucket_if_not_exist(provider, bucket_name, dir_name):
    logger.info(f"Creating bucket '{bucket_name}' in cloud provider '{provider}' with folder '{dir_name}'...")
    if provider == "s3":
        import boto3
        s3_client = boto3.client('s3')
        # s3_client.create_bucket(Bucket=bucket_name)
        s3_client.put_object(Bucket=bucket_name, Key=f"{dir_name}/")
        logger.info(f"Bucket '{bucket_name}' created in AWS S3.")
    elif provider == "gs":
        from google.cloud import storage
        """Create a folder in a Google Cloud Storage bucket if it does not exist."""
        storage_client = storage.Client()
        bucket = storage_client.bucket(bucket_name)

        # Ensure the folder name ends with a '/'
        if not dir_name.endswith('/'):
            dir_name += '/'

        # Check if folder exists by trying to list objects with the folder prefix
        blobs = storage_client.list_blobs(bucket_name, prefix=dir_name, delimiter='/')
        exists = any(blob.name == dir_name for blob in blobs)

        if not exists:
            # Create an empty blob to simulate a folder
            blob = bucket.blob(dir_name)
            blob.upload_from_string('')
            print(f"Folder '{dir_name}' created.")
        else:
            print(f"Folder '{dir_name}' already exists.")
    else:
        raise ValueError(f"Unsupported cloud provider '{provider}'.")

class Dataset:
    """
    Create or load from a new dataset.
    """

    def __init__(
        self,
        name: str,
        path: str = None,
        replace_existing: bool = False,
        features: Dict[
            str, FeatureType
        ] = {},  # features to be stored {name: FeatureType}
        enable_feature_inference=True,  # whether additional features can be inferred
        episode_info_connector: DatabaseConnector = None,
        step_data_connector: DatabaseConnector = None,
        storage: Optional[str] = None,
    ) -> None:
        """

        Args:
            name (str): Name of this dataset. Used as the directory name when exporting.
            path (str): Required. Local path of where this dataset should be stored.
            features (optional Dict[str, FeatureType]): Description of `param1`.
            enable_feature_inference (bool): enable inferring additional FeatureTypes

        Example:
        ```
        >>> dataset = fog_x.Dataset('my_dataset', path='~/fog_x/my_dataset`)
        ```

        TODO:
            * is replace_existing actually used anywhere?
        """
        self.name = name
        path = os.path.expanduser(path).removesuffix("/")
        logger.info(f"Dataset path: {path}")
        self.path = path
        if path is None:
            raise ValueError("Path is required")
        # create the folder if path doesn't exist
        if self.path.startswith("/") and not os.path.exists(path):
            logger.info(f"Creating directory {path}")
            os.makedirs(path)

        self.replace_existing = replace_existing
        self.features = features
        self.enable_feature_inference = enable_feature_inference
        if episode_info_connector is None:
                episode_info_connector = DataFrameConnector(f"{path}")
        
        if step_data_connector is None:
            if self.path.startswith("/") and not os.path.exists(f"{path}/{name}"):
                os.makedirs(f"{path}/{name}")
            try: 
                step_data_connector = LazyFrameConnector(f"{path}/{name}")
            except:
                logger.info(f"Path does not exist. ({path}/{name})")
                cloud_provider = path[:2]
                bucket_name = path[5:]
                create_cloud_bucket_if_not_exist(cloud_provider, bucket_name, f"{name}/")
                step_data_connector = LazyFrameConnector(f"{path}/{name}") 
        self.db_manager = DatabaseManager(episode_info_connector, step_data_connector)
        self.db_manager.initialize_dataset(self.name, features)

        self.storage = storage
        self.obs_keys = []
        self.act_keys = []
        self.step_keys = []

    def new_episode(self, metadata: Optional[Dict[str, Any]] = None) -> Episode:
        """
        Create a new episode / trajectory.

        Returns:
            Episode

        TODO:
            * support multiple processes writing to the same episode
            * close the previous episode if not closed
        """
        return Episode(
            metadata=metadata,
            features=self.features,
            enable_feature_inference=self.enable_feature_inference,
            db_manager=self.db_manager,
        )

    def _get_tf_feature_dicts(
        self, obs_keys: List[str], act_keys: List[str], step_keys: List[str]
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        """
        Get the tensorflow feature dictionaries.
        """
        observation_tf_dict = {}
        action_tf_dict = {}
        step_tf_dict = {}

        for k in obs_keys:
            observation_tf_dict[k] = self.features[k].to_tf_feature_type()

        for k in act_keys:
            action_tf_dict[k] = self.features[k].to_tf_feature_type()

        for k in step_keys:
            step_tf_dict[k] = self.features[k].to_tf_feature_type()

        return observation_tf_dict, action_tf_dict, step_tf_dict

    def export(
        self,
        export_path: Optional[str] = None,
        format: str = "rtx",
        max_episodes_per_file: int = 1,
        version: str = "0.0.1",
        obs_keys=[],
        act_keys=[],
        step_keys=[],
    ) -> None:
        """
        Export the dataset.

        Args:
            export_path (optional str): location of exported data. Uses dataset.path/export by default.
            format (str): Supported formats are `rtx`, `open-x`, and `rlds`.
        """
        if format == "rtx" or format == "open-x" or format == "rlds":
            self.export_rtx(export_path, max_episodes_per_file, version, obs_keys, act_keys, step_keys)
        else:
            raise ValueError("Unsupported export format")

    def export_rtx(        
        self,
        export_path: Optional[str] = None,
        max_episodes_per_file: int = 1,
        version: str = "0.0.1",
        obs_keys=[],
        act_keys=[],
        step_keys=[]
    ):
        if export_path == None:
            export_path = self.path + "/export"
        if not os.path.exists(export_path):
            os.makedirs(export_path)

        import dm_env
        import tensorflow as tf
        import tensorflow_datasets as tfds
        from envlogger import step_data
        from tensorflow_datasets.core.features import Tensor

        from fog_x.rlds.writer import CloudBackendWriter

        self.obs_keys += obs_keys
        self.act_keys += act_keys
        self.step_keys += step_keys

        (
            observation_tf_dict,
            action_tf_dict,
            step_tf_dict,
        ) = self._get_tf_feature_dicts(
            self.obs_keys,
            self.act_keys,
            self.step_keys,
        )

        logger.info("Exporting dataset as RT-X format")
        logger.info(f"Observation keys: {observation_tf_dict}")
        logger.info(f"Action keys: {action_tf_dict}")
        logger.info(f"Step keys: {step_tf_dict}")

        # generate tensorflow configuration file
        ds_config = tfds.rlds.rlds_base.DatasetConfig(
            name=self.name,
            description="",
            homepage="",
            citation="",
            version=tfds.core.Version("0.0.1"),
            release_notes={
                "0.0.1": "Initial release.",
            },
            observation_info=observation_tf_dict,
            action_info=action_tf_dict,
            reward_info=(
                step_tf_dict["reward"]
                if "reward" in step_tf_dict
                else Tensor(shape=(), dtype=tf.float32)
            ),
            discount_info=(
                step_tf_dict["discount"]
                if "discount" in step_tf_dict
                else Tensor(shape=(), dtype=tf.float32)
            ),
        )

        ds_identity = tfds.core.dataset_info.DatasetIdentity(
            name=ds_config.name,
            version=tfds.core.Version(version),
            data_dir=export_path,
            module_name="",
        )
        writer = CloudBackendWriter(
            data_directory=export_path,
            ds_config=ds_config,
            ds_identity=ds_identity,
            max_episodes_per_file=max_episodes_per_file,
        )

        # export the dataset
        episodes = self.get_episodes_from_metadata()
        for episode in episodes:
            steps = episode.collect().rows(named=True)
            for i in range(len(steps)):
                step = steps[i]
                observationd = {}
                actiond = {}
                stepd = {}
                for k, v in step.items():
                    # logger.info(f"key: {k}")
                    if k not in self.features:
                        if k != "episode_id" and k != "Timestamp":
                            logger.info(
                                f"Feature {k} not found in the dataset features."
                            )
                        continue
                    feature_spec = self.features[k].to_tf_feature_type()
                    if (
                        isinstance(feature_spec, tfds.core.features.Tensor)
                        and feature_spec.shape != ()
                    ):
                        # reverse the process
                        value = np.load(io.BytesIO(v)).astype(
                            feature_spec.np_dtype
                        )
                    elif (
                        isinstance(feature_spec, tfds.core.features.Tensor)
                        and feature_spec.shape == ()
                    ):
                        value = np.array(v, dtype=feature_spec.np_dtype)
                    elif isinstance(
                        feature_spec, tfds.core.features.Image
                    ):
                        value = np.load(io.BytesIO(v)).astype(
                            feature_spec.np_dtype
                        )
                    else:
                        value = v

                    if k in self.obs_keys:
                        observationd[k] = value
                    elif k in self.act_keys:
                        actiond[k] = value
                    else:
                        stepd[k] = value

                # logger.info(
                #     f"Step: {stepd}"
                #     f"Observation: {observationd}"
                #     f"Action: {actiond}"
                # )
                timestep = dm_env.TimeStep(
                    step_type=dm_env.StepType.FIRST,
                    reward=np.float32(
                        0.0
                    ),  # stepd["reward"] if "reward" in step else np.float32(0.0),
                    discount=np.float32(
                        0.0
                    ),  # stepd["discount"] if "discount" in step else np.float32(0.0),
                    observation=observationd,
                )
                stepdata = step_data.StepData(
                    timestep=timestep, action=actiond, custom_data=None
                )
                if i < len(steps) - 1:
                    writer._record_step(stepdata, is_new_episode=False)
                else:
                    writer._record_step(stepdata, is_new_episode=True)


    def load_rtx_episodes(
        self,
        name: str,
        split: Optional[str] = None,
        additional_metadata: Optional[Dict[str, Any]] = dict(),
    ):
        """
        Load robot data from Tensorflow Datasets.

        Args:
            name (str): Name of RT-X episodes, which can be found at [Tensorflow Datasets](https://www.tensorflow.org/datasets/catalog) under the Robotics category
            split (optional str): the portion of data to load, see [Tensorflow Split API](https://www.tensorflow.org/datasets/splits)
            additional_metadata (optional Dict[str, Any]): additional metadata to be associated with the loaded episodes

        Example:
            ```
            >>> dataset.load_rtx_episodes(name="berkeley_autolab_ur5)
            >>> dataset.load_rtx_episodes(name="berkeley_autolab_ur5", split="train[:10]", additional_metadata={"data_collector": "Alice", "custom_tag": "sample"})
            ```
        """

        # this is only required if rtx format is used
        import tensorflow_datasets as tfds

        from fog_x.rlds.utils import dataset2path

        b = tfds.builder_from_directory(builder_dir=dataset2path(name))

        ds = b.as_dataset(split=split)

        data_type = b.info.features["steps"]

        for tf_episode in ds:
            logger.info(tf_episode)
            fog_episode = self.new_episode(
                    metadata=additional_metadata,
                )
            for step in tf_episode["steps"]:
                ret = self._load_rtx_step_data_from_tf_step(
                    step, data_type, 
                )
                for r in ret:
                    fog_episode.add(**r)

            fog_episode.close()


    def _prepare_rtx_metadata(
        self,
        name: str,
        export_path: Optional[str] = None,
        sample_size = 20,
        shuffle = False,
        seed = 42,
    ):

        # this is only required if rtx format is used
        import tensorflow_datasets as tfds
        from fog_x.rlds.utils import dataset2path
        import cv2

        b = tfds.builder_from_directory(builder_dir=dataset2path(name))
        ds = b.as_dataset(split="all")
        if shuffle:
            ds = ds.shuffle(sample_size, seed=seed)
        data_type = b.info.features["steps"]
        counter = 0

        if export_path == None:
            export_path = self.path + "/" + self.name + "_viz"
        if not os.path.exists(export_path):
            os.makedirs(export_path)

        
        for tf_episode in ds:
            video_writers = {}

            additional_metadata = {
                "load_from": name,
                "load_index": f"all, {shuffle}, {seed}, {counter}",
            }
            
            logger.info(tf_episode)
            fog_episode = self.new_episode()
            
            for step in tf_episode["steps"]:
                ret = self._load_rtx_step_data_from_tf_step(
                    step, data_type, 
                )

                for r in ret:
                    feature_name = r["feature"]
                    if "image" in feature_name and "depth" not in feature_name:
                        image = np.load(io.BytesIO(r["value"]))

                        # convert from RGB to BGR
                        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                
                        if feature_name not in video_writers:
                            
                            output_filename = f"{self.name}_{counter}_{feature_name}"
                            tmp_vid_output_path = f"/tmp/{output_filename}.mp4"
                            output_path = f"{export_path}/{output_filename}"

                            frame_size = (image.shape[1], image.shape[0])

                            # save the initial image
                            cv2.imwrite(f"{output_path}.jpg", image)
                            # save the video
                            video_writers[feature_name] = cv2.VideoWriter(
                                tmp_vid_output_path,
                                cv2.VideoWriter_fourcc(*"mp4v"),
                                10,
                                frame_size
                            )
                            

                        video_writers[r["feature"]].write(image)

                    if "instruction" in r["feature"]:
                        natural_language_instruction = r["value"].decode("utf-8")
                        additional_metadata["natural_language_instruction"] = natural_language_instruction

                    r["metadata_only"] = True
                    fog_episode.add(**r)
            
            for feature_name, video_writer in video_writers.items():
                video_writer.release()
                # need to convert to h264 to properly display over chrome / vscode 
                output_filename = f"{self.name}_{counter}_{feature_name}"
                tmp_vid_output_path = f"/tmp/{output_filename}.mp4"
                vid_output_path = f"{export_path}/{output_filename}.mp4"
                convert_to_h264(tmp_vid_output_path, vid_output_path)
                additional_metadata[f"video_path_{feature_name}"] = output_filename
                if os.path.isfile(tmp_vid_output_path):
                    os.remove(tmp_vid_output_path)
                
            video_writers = {}
            fog_episode.close(save_data = False, additional_metadata = additional_metadata)
            counter += 1
            if counter > sample_size:
                break

    def _load_rtx_step_data_from_tf_step(
            self, 
            step: Dict[str, Any],
            data_type: Dict[str, Any] = {},
    ): 
        from tensorflow_datasets.core.features import (
            FeaturesDict,
            Image,
            Scalar,
            Tensor,
            Text,
        )
        ret = []

        for k, v in step.items():
            # logger.info(f"k {k} , v {v}")
            if isinstance(v, dict): #and (k == "observation" or k == "action"):
                for k2, v2 in v.items():
                    # TODO: abstract this to feature.py

                    if (
                        isinstance(data_type[k][k2], Tensor)
                        and data_type[k][k2].shape != ()
                    ):
                        memfile = io.BytesIO()
                        np.save(memfile, v2.numpy())
                        value = memfile.getvalue()
                    elif isinstance(data_type[k][k2], Image):
                        memfile = io.BytesIO()
                        np.save(memfile, v2.numpy())
                        value = memfile.getvalue()
                    else:
                        value = v2.numpy()
                    
                    ret.append(
                        {
                            "feature": str(k2),
                            "value": value,
                            "feature_type": FeatureType(
                                tf_feature_spec=data_type[k][k2]
                            ),
                        }
                    )
                    # fog_episode.add(
                    #     feature=str(k2),
                    #     value=value,
                    #     feature_type=FeatureType(
                    #         tf_feature_spec=data_type[k][k2]
                    #     ),
                    # )
                    if k == "observation":
                        self.obs_keys.append(k2)
                    elif k == "action":
                        self.act_keys.append(k2)
            else:
                # fog_episode.add(
                #     feature=str(k),
                #     value=v.numpy(),
                #     feature_type=FeatureType(tf_feature_spec=data_type[k]),
                # )
                ret.append(
                    {
                        "feature": str(k),
                        "value": v.numpy(),
                        "feature_type": FeatureType(
                            tf_feature_spec=data_type[k]
                        ),
                    }
                )
                self.step_keys.append(k)
        return ret
        

    def get_episode_info(self) -> pandas.DataFrame:
        """
        Returns:
            metadata of all episodes as `pandas.DataFrame`
        """
        return self.db_manager.get_episode_info_table()

    def get_step_data(self) -> polars.LazyFrame:
        """
        Returns:
            step data of all episodes
        """
        return self.db_manager.get_step_table_all()

    def get_step_data_by_episode_ids(
        self, episode_ids: List[int], as_lazy_frame=True
    ):
        """
        Args:
            episode_ids (List[int]): list of episode ids
            as_lazy_frame (bool): whether to return polars.LazyFrame or polars.DataFrame

        Returns:
            step data of each episode
        """
        episodes = []
        for episode_id in episode_ids:
            if episode_id == None:
                continue
            if as_lazy_frame:
                episodes.append(self.db_manager.get_step_table(episode_id))
            else:
                episodes.append(self.db_manager.get_step_table(episode_id).collect())
        return episodes

    def read_by(self, episode_info: Any = None) -> List[polars.LazyFrame]:
        """
        To be used with `Dataset.get_episode_info`.

        Args:
            episode_info (pandas.DataFrame): episode metadata information to determine which episodes to read

        Returns:
            episodes filtered by `episode_info`
        """
        episode_ids = list(episode_info["episode_id"])
        logger.info(f"Reading episodes as order: {episode_ids}")
        episodes = []
        for episode_id in episode_ids:
            if episode_id == None:
                continue
            episodes.append(self.db_manager.get_step_table(episode_id))
        return episodes

    def get_episodes_from_metadata(self, metadata: Any = None):
        # Assume we use get_metadata_as_pandas_df to retrieve episodes metadata
        if metadata is None:
            metadata_df = self.get_episode_info()
        else:
            metadata_df = metadata
        episodes = self.read_by(metadata_df)
        return episodes

    def pytorch_dataset_builder(self, metadata=None, **kwargs):
        """
        Used for loading current dataset as a PyTorch dataset.
        To be used with `torch.utils.data.DataLoader`.
        """

        import torch
        from torch.utils.data import Dataset
        episodes = self.get_episodes_from_metadata(metadata)

        # Initialize the PyTorch dataset with the episodes and features
        pytorch_dataset = PyTorchDataset(episodes, self.features)

        return pytorch_dataset

    def get_as_huggingface_dataset(self):
        """
        Load current dataset as a HuggingFace dataset.

        TODO:
            * currently the support for huggingg face dataset is limited.
                it only shows its capability of easily returning a hf dataset
            * add features from the episode metadata
            * allow selecting episodes based on queries.
                doing so requires creating a new copy of the dataset on disk
        """
        import datasets

        dataset_path = self.path + "/" + self.name
        parquet_files = [
            os.path.join(dataset_path, f) for f in os.listdir(dataset_path)
        ]

        hf_dataset = datasets.load_dataset("parquet", data_files=parquet_files)
        return hf_dataset
    
class PyTorchDataset(Dataset):
    def __init__(self, episodes, features):
        """
        Initialize the dataset with the episodes and features.
        :param episodes: A list of episodes loaded from the database.
        :param features: A dictionary of features to be included in the dataset.
        """
        self.episodes = episodes
        self.features = features

    def __len__(self):
        """
        Return the total number of episodes in the dataset.
        """
        return len(self.episodes)

    def __getitem__(self, idx):
        """
        Retrieve the idx-th episode from the dataset.
        Depending on the structure, you may need to process the episode
        and its features here.
        """
        print("Retrieving episode at index", idx)
        episode = self.episodes[idx].collect().to_pandas()
        # Process the episode and its features here
        # For simplicity, let's assume we're just returning the episode
        return episode



import logging
from typing import Any, List, Optional, Tuple

import numpy as np
from sqlalchemy import Float, Integer, LargeBinary, String

from fog_x.database.utils import type_np2sql, type_py2sql

logger = logging.getLogger(__name__)


SUPPORTED_DTYPES = [
    "null",
    "bool",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "float16",
    "float32",
    "float64",
    "timestamp(s)",
    "timestamp(ms)",
    "timestamp(us)",
    "timestamp(ns)",
    "timestamp(s, tz)",
    "timestamp(ms, tz)",
    "timestamp(us, tz)",
    "timestamp(ns, tz)",
    "binary",
    "large_binary",
    "string",
    "str",
    "large_string",
]


class FeatureType:
    """
    class for feature definition and conversions
    """

    def __init__(
        self,
        dtype: Optional[str] = None,
        shape: Any = None,
        tf_feature_spec=None,
        data=None,
    ) -> None:
        # scalar: (), vector: (n,), matrix: (n,m)
        if data is not None:
            self.from_data(data)
        elif tf_feature_spec is not None:
            self.from_tf_feature_type(tf_feature_spec)
        elif dtype is not None:
            self._set(dtype, shape)
        else:
            raise ValueError("Either dtype or data must be provided")

    def __str__(self):
        return f"dtype={self.dtype}, shape={self.shape})"

    def __repr__(self):
        return self.__str__()

    def _set(self, dtype: str, shape: Any):
        if dtype == "double":  # fix inferred type
            dtype = "float64"
        if dtype == "float":  # fix inferred type
            dtype = "float32"
        if dtype not in SUPPORTED_DTYPES:
            raise ValueError(f"Unsupported dtype: {dtype}")
        if shape is not None and not isinstance(shape, tuple):
            raise ValueError(f"Shape must be a tuple: {shape}")
        self.dtype = dtype
        self.shape = shape

    def from_tf_feature_type(self, tf_feature_spec):
        """
        Convert from tf feature
        """
        logger.debug(f"tf_feature_spec: {tf_feature_spec}")
        from tensorflow_datasets.core.features import (
            FeaturesDict,
            Image,
            Scalar,
            Tensor,
            Text,
        )

        if isinstance(tf_feature_spec, Tensor):
            shape = tf_feature_spec.shape
            dtype = tf_feature_spec.dtype.name
        elif isinstance(tf_feature_spec, Image):
            shape = tf_feature_spec.shape
            dtype = tf_feature_spec.np_dtype
            # TODO: currently images are not handled efficiently
        elif isinstance(tf_feature_spec, Scalar):
            shape = ()
            dtype = tf_feature_spec.dtype.name
        elif isinstance(tf_feature_spec, Text):
            shape = ()
            dtype = "string"
        else:
            raise ValueError(
                f"Unsupported conversion from tf feature: {tf_feature_spec}"
            )
        self._set(str(dtype), shape)
        return self

    def from_data(self, data: Any):
        """
        Infer feature type from the provided data.
        """
        if isinstance(data, np.ndarray):
            self._set(data.dtype.name, data.shape)
        elif isinstance(data, list):
            dtype = type(data[0]).__name__
            shape = (len(data),)
            self._set(dtype.name, shape)
        else:
            dtype = type(data).__name__
            shape = ()
            self._set(dtype, shape)
        return self

    def to_tf_feature_type(self):
        """
        Convert to tf feature
        """
        import tensorflow as tf
        from tensorflow_datasets.core.features import (
            FeaturesDict,
            Image,
            Scalar,
            Tensor,
            Text,
        )

        str_dtype_to_tf_dtype = {
            "int8": tf.int8,
            "int16": tf.int16,
            "int32": tf.int32,
            "int64": tf.int64,
            "uint8": tf.uint8,
            "uint16": tf.uint16,
            "uint32": tf.uint32,
            "uint64": tf.uint64,
            "float16": tf.float16,
            "float32": tf.float32,
            "float64": tf.float64,
            "string": tf.string,
            "str": tf.string,
            "bool": tf.bool,
        }
        tf_detype = str_dtype_to_tf_dtype[self.dtype]
        if len(self.shape) == 0:
            if self.dtype == "string":
                return Text()
            else:
                return Scalar(dtype=tf_detype)
        elif len(self.shape) >= 1:
            return Tensor(shape=self.shape, dtype=tf_detype)
        else:
            raise ValueError(f"Unsupported conversion to tf feature: {self}")

    def to_sql_type(self):
        """
        Convert to sql type
        """
        if self.is_np:
            return LargeBinary
        else:
            try:
                return type_np2sql(self.dtype)
            except:
                return LargeBinary

    def to_pld_storage_type(self):
        if len(self.shape) == 0:
            if self.dtype == "string":
                return "large_binary"  # TODO: better representation of strings
            else:
                return self.dtype
        else:
            return "large_binary"



# Copyright 2022 The Regents of the University of California (Regents)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright ©2022. The Regents of the University of California (Regents).
# All Rights Reserved. Permission to use, copy, modify, and distribute this
# software and its documentation for educational, research, and not-for-profit
# purposes, without fee and without a signed licensing agreement, is hereby
# granted, provided that the above copyright notice, this paragraph and the
# following two paragraphs appear in all copies, modifications, and
# distributions. Contact The Office of Technology Licensing, UC Berkeley, 2150
# Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# otl@berkeley.edu, http://ipira.berkeley.edu/industry-info for commercial
# licensing opportunities. IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
# DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE. REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
# PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO PROVIDE
# MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.


# coding=utf-8
# Copyright 2023 DeepMind Technologies Limited..
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFDS backend for Envlogger."""
import dataclasses
from collections import ChainMap
from typing import Any, Dict, List, Optional

import tensorflow_datasets as tfds
from envlogger import step_data
from envlogger.backends import backend_writer, rlds_utils

DatasetConfig = tfds.rlds.rlds_base.DatasetConfig

import logging

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Episode(object):
    """Episode that is being constructed."""

    prev_step: step_data.StepData
    steps: Optional[List[rlds_utils.Step]] = None
    metadata: Optional[Dict[str, Any]] = None

    def add_step(self, step: step_data.StepData) -> None:
        rlds_step = rlds_utils.to_rlds_step(self.prev_step, step)
        if self.steps is None:
            self.steps = []
        self.steps.append(rlds_step)
        self.prev_step = step

    def get_rlds_episode(self) -> Dict[str, Any]:
        last_step = rlds_utils.to_rlds_step(self.prev_step, None)
        if self.steps is None:
            self.steps = []
        if self.metadata is None:
            self.metadata = {}

        return {"steps": self.steps + [last_step], **self.metadata}


class CloudBackendWriter(backend_writer.BackendWriter):
    """Backend that writes trajectory data in TFDS format (and RLDS structure)."""

    def __init__(
        self,
        data_directory: str,
        ds_config: tfds.rlds.rlds_base.DatasetConfig,
        ds_identity: tfds.core.dataset_info.DatasetIdentity,
        max_episodes_per_file: int = 1,
        split_name: Optional[str] = None,
        version: str = "0.0.1",
        store_ds_metadata: bool = False,
        **base_kwargs
    ):
        """Constructor.

        Args:
          data_directory: Directory to store the data
          ds_config: Dataset Configuration.
          max_episodes_per_file: Number of episodes to store per shard.
          split_name: Name to be used by the split. If None, 'train' will be used.
          version: version (major.minor.patch) of the dataset.
          store_ds_metadata: if False, it won't store the dataset level
            metadata.
          **base_kwargs: arguments for the base class.
        """
        super().__init__(**base_kwargs)
        if not split_name:
            split_name = "train"
        if store_ds_metadata:
            metadata = self._metadata
        else:
            metadata = None
        self._data_directory = data_directory
        self._ds_info = tfds.rlds.rlds_base.build_info(
            ds_config, ds_identity, metadata
        )
        self._ds_info.set_file_format("tfrecord")

        self._current_episode = None

        self._sequential_writer = tfds.core.SequentialWriter(
            self._ds_info, max_episodes_per_file
        )
        self._split_name = split_name
        self._sequential_writer.initialize_splits([split_name])
        logging.info("self._data_directory: %r", self._data_directory)

    def _write_and_reset_episode(self):
        if self._current_episode is not None:
            self._sequential_writer.add_examples(
                {self._split_name: [self._current_episode.get_rlds_episode()]}
            )
            self._current_episode = None

    def _record_step(
        self, data: step_data.StepData, is_new_episode: bool
    ) -> None:
        """Stores RLDS steps in TFDS format."""

        if is_new_episode:
            self._write_and_reset_episode()

        if self._current_episode is None:
            self._current_episode = Episode(prev_step=data)
        else:
            self._current_episode.add_step(data)

    def set_episode_metadata(self, data: Dict[str, Any]) -> None:
        self._current_episode.metadata = data

    def close(self) -> None:
        logging.info(
            "Deleting the backend with data_dir: %r", self._data_directory
        )
        self._write_and_reset_episode()
        self._sequential_writer.close_all()
        logging.info(
            "Done deleting the backend with data_dir: %r", self._data_directory
        )



import logging
from ast import literal_eval
from typing import Any, Dict, List, Optional
import numpy as np
from fog_x.database.polars_connector import PolarsConnector
from fog_x.feature import FeatureType
import io
logger = logging.getLogger(__name__)


class DatabaseManager:
    """
    DatabaseManager class for adding and querying data from the database
    """

    def __init__(
        self,
        episode_info_connector: PolarsConnector,
        step_data_connector: PolarsConnector,
        required_stats: List[str] = ["count"],
    ):
        self.episode_info_connector = episode_info_connector
        self.step_data_connector = step_data_connector
        self.dataset_name: Optional[str] = None
        self.current_episode_id: Optional[int] = None
        self.features: Dict[str, FeatureType] = {}
        self.current_episode_id = -1
        self.episode_metadata = {}
        self.required_stats = required_stats

    def initialize_dataset(
        self, dataset_name: str, features: Dict[str, FeatureType]
    ):
        self.dataset_name = dataset_name
        self.features = features
        self.episode_info_connector.load_tables([self.dataset_name])

        # tables = self.episode_info_connector.list_tables()

        episode_info_table = None
        episode_info_table = self.get_episode_info_table()

        logger.debug(f"Episode Info Table: {episode_info_table}")
        if episode_info_table is not None:
            self.current_episode_id = len(episode_info_table)
            for row in episode_info_table.iter_rows(named=True):
                if None in row.values():
                    continue
                for key, value in row.items():
                    if key.endswith("type"):
                        feature_name = key.replace("feature_", "").replace(
                            "_type", ""
                        )
                        feature_type = FeatureType(
                            dtype=value,
                            shape=literal_eval(
                                row[f"feature_{feature_name}_shape"]
                            ),
                        )
                        self.features[feature_name] = feature_type
                        logger.debug(
                            f"Loaded Features: {feature_name} with type {feature_type}"
                        )
                break

            self.step_data_connector.load_tables(
                [episode_id for episode_id in range(len(episode_info_table))],
                [
                    f"{self.dataset_name}_{episode_id}"
                    for episode_id in range(len(episode_info_table))
                ],
            )
            # if tables and dataset_name in tables:
            #     logger.debug("Database is not empty and already initialized")
            #     # get features from the table and update the features
            #     metadata = self.get_metadata_table()
            #     for _, row in metadata.iterrows():
            #         feature_type = FeatureType(
            #             dtype=row["Type"], shape=row["Shape"]
            #         )
            #         self.features[row["Feature"]] = feature_type
            #     logger.debug(
            #         f"Loaded Features: {self.features} with type {feature_type}"
            #     )
        else:
            self.episode_info_connector.create_table(
                dataset_name,
            )
            self.episode_info_connector.add_column(
                dataset_name,
                "episode_id",
                "int64",
            )
            self.episode_info_connector.add_column(
                dataset_name,
                "Finished",
                "bool",
            )
            self.current_episode_id = 0
            logger.debug("Database initialized")

    def initialize_episode(
        self,
        additional_metadata: Optional[Dict[str, Any]] = None,
    ) -> int:
        if self.dataset_name is None:
            raise ValueError("Dataset not initialized")
        if additional_metadata is None:
            additional_metadata = {}
        for key, value in additional_metadata.items():
            self.episode_metadata[key] = value

        self.episode_metadata["episode_id"] = self.current_episode_id
        self.episode_metadata["Finished"] = False
        logger.debug(
            f"Initializing episode for dataset {self.dataset_name} with metadata {self.episode_metadata}"
        )
        for metadata_key in self.episode_metadata.keys():
            logger.debug(f"Adding metadata key {metadata_key} to the database")
            self.episode_info_connector.add_column(
                self.dataset_name,
                metadata_key,
                "str",  # TODO: support more types
            )

        # insert episode information to the database
        self.episode_info_connector.insert_data(
            self.dataset_name,
            self.episode_metadata,
        )

        # create tables for each feature
        for feature_name, feature_type in self.features.items():
            self._initialize_feature(feature_name, feature_type)

        return self.current_episode_id

    def _initialize_feature(
        self, feature_name: str, feature_type: FeatureType
    ):
        # create a table for the feature
        # TODO: need to make the timestamp type as TIMESTAMPTZ
        self.step_data_connector.create_table(
            self._get_feature_table_name(feature_name),
        )
        # {"Timestamp": "int64", feature_name: feature_type.to_sql_type()}

        self.step_data_connector.add_column(
            self._get_feature_table_name(feature_name),
            "episode_id",
            "int64",
        )
        self.step_data_connector.add_column(
            self._get_feature_table_name(feature_name),
            "Timestamp",
            "int64",
        )
        logger.debug(
            f"Adding feature {feature_name} to the database with type {feature_type.to_pld_storage_type()}"
        )
        self.step_data_connector.add_column(
            self._get_feature_table_name(feature_name),
            feature_name,
            feature_type.to_pld_storage_type(),
        )

        if feature_type is None:
            logger.error(f"Feature type not provided for {feature_name}")
            self.features[feature_name] = FeatureType()
        else:
            self.features[feature_name] = feature_type

        # update dataset metadata with the feature information
        self.episode_info_connector.add_column(
            self.dataset_name,
            f"feature_{feature_name}_type",
            "str",
        )
        self.episode_info_connector.add_column(
            self.dataset_name,
            f"feature_{feature_name}_shape",
            "str",
        )
        self.episode_info_connector.update_data(
            table_name=self.dataset_name,
            index=self.current_episode_id,
            data={
                f"feature_{feature_name}_type": str(
                    self.features[feature_name].dtype
                ),
                f"feature_{feature_name}_shape": str(
                    self.features[feature_name].shape
                ),
            },
        )
        for stat in self.required_stats:
            self.episode_info_connector.add_column(
                self.dataset_name,
                f"{feature_name}_{stat}",
                "float64",
            )

    def add(
        self,
        feature_name: str,
        value: Any,
        timestamp: int,
        feature_type: Optional[FeatureType] = None,
        metadata_only = False,
    ):
        if feature_name not in self.features.keys():
            logger.warning(
                f"Feature {feature_name} not in the list of features"
            )
            if feature_type is None:

                feature_type = FeatureType(data=value)
                logger.warn(
                    f"feature type not provided, inferring from data type {feature_type}"
                )
            self._initialize_feature(feature_name, feature_type)

        if not metadata_only:
            # TODO: any conversion is needed 
            if isinstance(value, np.ndarray):
                logger.debug(f"Converting numpy array to bytes")
                memfile = io.BytesIO()
                np.save(memfile, value)
                value = memfile.getvalue()
            # insert data into the table
            self.step_data_connector.insert_data(
                self._get_feature_table_name(feature_name),
                {
                    "episode_id": self.current_episode_id,
                    "Timestamp": timestamp,
                    feature_name: value,
                },
            )

    def compact(self):
        # create a table for the compacted data
        # iterate through all the features and get the data
        table_names = [
            self._get_feature_table_name(feature_name)
            for feature_name in self.features.keys()
        ]
        logger.debug(f"Compacting and Merging tables: {table_names}")

        self.step_data_connector.merge_tables_with_timestamp(
            table_names,
            f"{self.dataset_name}_{self.current_episode_id}",
            clear_feature_tables = True,
        )

        self.step_data_connector.remove_tables(table_names)

    # def get_table(
    #     self, table_name: Optional[str] = None, format: str = "pandas"
    # ):
    #     if table_name is None:
    #         table_name = self.dataset_name
    #     if table_name is None:
    #         raise ValueError("Dataset name not provided")
    #     return self.db_connector.select_table(table_name)

    def get_episode_info_table(self):
        return self.episode_info_connector.select_table(self.dataset_name)

    def get_step_table(self, episode_id):
        if episode_id is None:
            raise ValueError("Episode not initialized")
        return self.step_data_connector.select_table(
            f"{self.dataset_name}_{episode_id}",
        )

    def get_step_table_all(self, reload=False):
        return self.step_data_connector.get_dataset_table(reload)

    def _get_feature_table_name(self, feature_name):
        if self.dataset_name is None:
            raise ValueError("Dataset not initialized")
        if self.current_episode_id is None:
            raise ValueError("Episode not initialized")
        table_name = (
            f"{self.dataset_name}_{self.current_episode_id}_{feature_name}"
        )
        return table_name

    def close(self,  save_data=True, save_metadata=True, additional_metadata = None):

        if additional_metadata is None:
            additional_metadata = {}
        logger.info(f"Closing the episode with metadata {additional_metadata}")
        additional_metadata["Finished"] = True
        if save_data:
            self.compact()
            compacted_table = self.step_data_connector.select_table(
                f"{self.dataset_name}_{self.current_episode_id}"
            )
            for feature_name in self.features.keys():
                if "count" in self.required_stats:
                    additional_metadata[f"{feature_name}_count"] = compacted_table[
                        feature_name
                    ].count()
                if "mean" in self.required_stats:
                    additional_metadata[f"{feature_name}_mean"] = compacted_table[
                        feature_name
                    ].mean()
                if "max" in self.required_stats:
                    additional_metadata[f"{feature_name}_max"] = compacted_table[
                        feature_name
                    ].max()
                if "min" in self.required_stats:
                    additional_metadata[f"{feature_name}_min"] = compacted_table[
                        feature_name
                    ].min()
            self.step_data_connector.save_table(
                f"{self.dataset_name}_{self.current_episode_id}",
            )
            # TODO: this is a hack clear the old dataframe and load as a lazy frame  
            # TODO: future iteration: serve as cache
            self.step_data_connector.load_tables(
                [self.current_episode_id],
                [f"{self.dataset_name}_{self.current_episode_id}"],
            )
        else: 
            table_names = [
                self._get_feature_table_name(feature_name)
                for feature_name in self.features.keys()
            ]
            self.step_data_connector.remove_tables(table_names)

        for metadata_key in additional_metadata.keys():
            logger.debug(f"Adding metadata key {metadata_key} to the database")
            self.episode_info_connector.add_column(
                self.dataset_name,
                metadata_key,
                "str",  # TODO: support more types
            )
        # update the metadata field marking the episode as compacted
        self.episode_info_connector.update_data(
            self.dataset_name, self.current_episode_id, additional_metadata
        )

        self.episode_info_connector.save_table(
            f"{self.dataset_name}"
        )
        self.current_episode_id += 1
